import warnings

import requests
from packaging import version
from requests import exceptions


def _get_latest_version():
    response = requests.get("https://pypi.org/pypi/up42-py/json", timeout=2)
    response.raise_for_status()
    return version.parse(response.json()["info"]["version"])


def build_outdated_version_message(installed_version, latest_version):
    return f"You're using an outdated version of the UP42 Python SDK: v{installed_version}. A newer version is available: v{latest_version}.\nPlease upgrade to the latest version using **pip install --upgrade up42-py** or conda **conda update -c conda-forge up42-py**."  # pylint: disable=line-too-long # noqa: E501


def check_is_latest_version(
    installed_version: str,
    warn=warnings.warn,
    build_warning_message=build_outdated_version_message,
):
    try:
        latest_version = _get_latest_version()
        if version.Version(installed_version) < latest_version:
            warn(build_warning_message(installed_version, latest_version))
    except exceptions.HTTPError:
        pass



import datetime as dt
import enum
import itertools
from typing import List, Literal, Optional, Union

from up42 import asset, base, order, utils

logger = utils.get_logger(__name__)


class AllowedStatuses(enum.Enum):
    CREATED = "CREATED"
    BEING_PLACED = "BEING_PLACED"
    PLACED = "PLACED"
    PLACEMENT_FAILED = "PLACEMENT_FAILED"
    DELIVERY_INITIALIZATION_FAILED = "DELIVERY_INITIALIZATION_FAILED"
    BEING_FULFILLED = "BEING_FULFILLED"
    DOWNLOAD_FAILED = "DOWNLOAD_FAILED"
    DOWNLOADED = "DOWNLOADED"
    FULFILLED = "FULFILLED"
    FAILED_PERMANENTLY = "FAILED_PERMANENTLY"


OrderSortBy = Literal["createdAt", "updatedAt", "dataProvider", "type", "status"]
OrderType = Literal["TASKING", "ARCHIVE"]


class Storage:
    """
    The Storage class enables access to the UP42 storage. You can list
    your assets and orders within an UP42 workspace.

    Use the storage:
    ```python
    storage = up42.initialize_storage()
    ```
    """

    session = base.Session()
    workspace_id = base.WorkspaceId()
    pystac_client = base.StacClient()

    @utils.deprecation("Asset::all", "3.0.0")
    def get_assets(
        self,
        created_after: Optional[Union[str, dt.datetime]] = None,
        created_before: Optional[Union[str, dt.datetime]] = None,
        workspace_id: Optional[str] = None,
        collection_names: Optional[List[str]] = None,
        producer_names: Optional[List[str]] = None,
        tags: Optional[List[str]] = None,
        sources: Optional[List[str]] = None,
        search: Optional[str] = None,
        limit: Optional[int] = None,
        sortby: str = "createdAt",
        descending: bool = True,
        return_json: bool = False,
    ) -> Union[List[asset.Asset], List[dict]]:
        """
        Gets a list of assets in storage as [Asset](https://sdk.up42.com/structure/#asset) objects or in JSON format.

        Args:
            created_after: Search for assets created after the specified timestamp, in `"YYYY-MM-DD"` format.
            created_before: Search for assets created before the specified timestamp, in `"YYYY-MM-DD"` format.
            workspace_id: Search by the workspace ID.
            collection_names: Search for assets from any of the provided geospatial collections.
            producer_names: Search for assets from any of the provided producers.
            tags: Search for assets with any of the provided tags.
            sources: Search for assets from any of the provided sources.\
                The allowed values: `"ARCHIVE"`, `"TASKING"`, `"USER"`.
            search: Search for assets that contain the provided search query in their name, title, or order ID.
            limit: The number of results on a results page.
            sortby: The property to sort by.
            descending: The sorting order: <ul><li>`true` — descending</li><li>`false` — ascending</li></ul>
            return_json: If `true`, returns a JSON dictionary.\
                If `false`, returns a list of [Asset](https://sdk.up42.com/structure/#functionality_1) objects.

        Returns:
            A list of Asset objects.
        """
        assets = list(
            itertools.islice(
                asset.Asset.all(
                    created_after=created_after,
                    created_before=created_before,
                    workspace_id=workspace_id,
                    collection_names=collection_names,
                    producer_names=producer_names,
                    tags=tags,
                    sources=sources,
                    search=search,
                    sort_by=utils.SortingField(sortby, not descending),
                ),
                limit,
            )
        )
        if return_json:
            return [a.info for a in assets]
        else:
            return assets

    @utils.deprecation("Order::all", "3.0.0")
    def get_orders(
        self,
        workspace_orders: bool = True,
        return_json: bool = False,
        limit: Optional[int] = None,
        sortby: OrderSortBy = "createdAt",
        descending: bool = True,
        order_type: Optional[OrderType] = None,
        # FIXME: should be AllowedStatuses instead of str
        statuses: Optional[List[str]] = None,
        name: Optional[str] = None,
        tags: Optional[List[str]] = None,
    ) -> Union[List[order.Order], List[dict]]:
        """
        Gets all orders in the account/workspace as Order objects or JSON.

        Args:
            workspace_orders: If set to True, only returns workspace orders. Otherwise, returns all account orders.
            return_json: If set to True, returns JSON object.
            limit: Optional, only return n first assets by sorting criteria and order.
                Optimal to select if your workspace contains many assets.
            sortby: The sorting criteria, one of "createdAt", "updatedAt", "status", "dataProvider", "type".
            descending: The sorting order, True for descending (default), False for ascending.
            order_type: Can be either "TASKING" or "ARCHIVE". Pass this param to filter orders based on order_type.
            statuses: Search for orders with any of the statuses provided.
            name: Search for orders that contain this string in their name.
            tags: Search for orders with any of the provided tags.

        Returns:
            Order objects in the workspace or alternatively JSON info of the orders.
        """

        orders = list(
            itertools.islice(
                order.Order.all(
                    workspace_id=self.workspace_id if workspace_orders else None,
                    display_name=name,
                    order_type=order_type,
                    tags=tags,
                    status=statuses,  # type: ignore
                    sort_by=utils.SortingField(sortby, not descending),
                ),
                limit,
            )
        )

        if return_json:
            return [o.info for o in orders]
        else:
            return orders



DOMAIN = "com"
REGION = "eu"


def endpoint(path: str):
    """Gets endpoint url based on its path."""
    if REGION == "eu":
        return f"https://api.up42.{DOMAIN}{path}"
    elif REGION == "sa":
        return f"https://api.sa.up42.{DOMAIN}{path}"
    raise UnsupportedRegion(f"Region {REGION} is not supported")


def token_endpoint():
    if REGION == "eu":
        return f"https://auth.up42.{DOMAIN}/realms/public/protocol/openid-connect/token"
    elif REGION == "sa":
        return f"https://auth.sa.up42.{DOMAIN}/realms/public/protocol/openid-connect/token"
    raise UnsupportedRegion(f"Region {REGION} is not supported")


def user_info_endpoint():
    if REGION == "eu":
        return f"https://auth.up42.{DOMAIN}/realms/public/protocol/openid-connect/userinfo"
    elif REGION == "sa":
        return f"https://auth.sa.up42.{DOMAIN}/realms/public/protocol/openid-connect/userinfo"
    raise UnsupportedRegion(f"Region {REGION} is not supported")


class UnsupportedRegion(ValueError):
    pass



import dataclasses as dc
import datetime as dt
import threading
from typing import Optional, Protocol

import requests

from up42.http import config, http_adapter

CLIENT_ID = "up42-sdk"


@dc.dataclass(eq=True, frozen=True)
class Token:
    access_token: str
    expires_on: dt.datetime

    @property
    def has_expired(self) -> bool:
        return self.expires_on <= dt.datetime.now()


class TokenRetriever(Protocol):
    def __call__(self, session: requests.Session, settings: config.TokenProviderSettings) -> Token:
        ...


class AccountTokenRetriever:
    def __init__(self, settings: config.AccountCredentialsSettings):
        self.username = settings.username
        self.password = settings.password

    def __call__(self, session: requests.Session, settings: config.TokenProviderSettings) -> Token:
        headers = {
            "Content-Type": "application/x-www-form-urlencoded",
        }
        body = {
            "grant_type": "password",
            "username": self.username,
            "password": self.password,
            "client_id": CLIENT_ID,
            "scope": "openid",
        }
        response = session.post(
            url=settings.token_url,
            data=body,
            headers=headers,
            timeout=settings.timeout,
        )
        if response.ok:
            token_data = response.json()
            access_token = token_data["access_token"]
            expires_on = dt.datetime.now() + dt.timedelta(seconds=token_data["expires_in"] - settings.expiry_offset)
            return Token(access_token=access_token, expires_on=expires_on)
        raise WrongCredentials


class Up42Auth(requests.auth.AuthBase):
    def __init__(
        self,
        retrieve: TokenRetriever,
        token_settings: config.TokenProviderSettings,
        create_adapter=http_adapter.create,
    ):
        self.token_settings = token_settings
        self.adapter = create_adapter(include_post=True)
        self.retrieve = retrieve
        self._token = self._fetch_token()
        self._lock = threading.Lock()

    def __call__(self, request: requests.PreparedRequest) -> requests.PreparedRequest:
        request.headers["Authorization"] = f"Bearer {self._access_token}"
        return request

    def _fetch_token(self):
        session = requests.Session()
        session.mount("https://", self.adapter)
        return self.retrieve(session, self.token_settings)

    @property
    def _access_token(self) -> Token:
        with self._lock:
            if self._token.has_expired:
                self._token = self._fetch_token()
        return self._token.access_token

    def __deepcopy__(self, memo: dict):
        # Pystac client deep copies the request modifier this class is used for.
        # We return the same instance to share authentication settings between
        # possible parallel calls to avoid multiple token generation.
        return self


def detect_settings(
    credentials: Optional[dict],
) -> Optional[config.CredentialsSettings]:
    if credentials:
        if all(credentials.values()):
            if credentials.keys() == {"username", "password"}:
                return config.AccountCredentialsSettings(**credentials)
            raise InvalidCredentials
        elif any(credentials.values()):
            raise IncompleteCredentials
    return None


def detect_retriever(settings: config.CredentialsSettings):
    if isinstance(settings, config.AccountCredentialsSettings):
        return AccountTokenRetriever(settings)
    raise UnsupportedSettings(f"Settings {settings} are not supported")


class InvalidCredentials(ValueError):
    pass


class IncompleteCredentials(ValueError):
    pass


class UnsupportedSettings(ValueError):
    pass


class WrongCredentials(ValueError):
    pass



import dataclasses
import datetime as dt
import pathlib
from typing import Iterator, List, Literal, Optional, Union, cast

import pystac
import pystac_client
import tenacity as tnc

from up42 import base, host, utils

logger = utils.get_logger(__name__)

MAX_ITEM = 50
LIMIT = 50

NOT_PROVIDED = object()
_retry = tnc.retry(
    stop=tnc.stop_after_attempt(5),
    wait=tnc.wait_exponential(multiplier=1),
    reraise=True,
)


class AssetSorting:
    name = utils.SortingField("name")
    title = utils.SortingField("title")
    size = utils.SortingField("size")
    created_at = utils.SortingField("createdAt")
    updated_at = utils.SortingField("updatedAt")
    source = utils.SortingField("source")
    collection_name = utils.SortingField("collectionName")
    producer_name = utils.SortingField("producerName")


@dataclasses.dataclass
class Asset:
    session = base.Session()
    stac_client = base.StacClient()
    id: str
    account_id: str
    created_at: str
    name: Optional[str]
    size: int
    updated_at: str
    workspace_id: str
    order_id: Optional[str]
    source: Optional[Literal["ARCHIVE", "TASKING", "PROCESSING"]]
    product_id: Optional[str]
    content_type: str
    producer_name: Optional[str]
    collection_name: Optional[str]
    geospatial_metadata_extraction_status: Optional[Literal["SUCCESSFUL", "FAILED", "IN_PROGRESS", "NOT_PROCESSED"]]
    title: Optional[str]
    tags: Optional[list[str]]
    info: dict

    @property
    @utils.deprecation(None, "3.0.0")
    def asset_id(self) -> str:
        return self.info["id"]

    @staticmethod
    def _from_metadata(metadata: dict) -> "Asset":
        return Asset(
            id=metadata["id"],
            account_id=metadata["accountId"],
            created_at=metadata["createdAt"],
            name=metadata.get("name"),
            size=metadata["size"],
            updated_at=metadata["updatedAt"],
            workspace_id=metadata["workspaceId"],
            order_id=metadata.get("orderId"),
            source=metadata.get("source"),
            product_id=metadata.get("productId"),
            content_type=metadata["contentType"],
            producer_name=metadata.get("producerName"),
            collection_name=metadata.get("collectionName"),
            geospatial_metadata_extraction_status=metadata.get("geospatialMetadataExtractionStatus"),
            title=metadata.get("title"),
            tags=metadata.get("tags"),
            info=metadata,
        )

    @classmethod
    @utils.deprecation(None, "3.0.0")
    def get(cls, asset_id: str) -> "Asset":
        url = host.endpoint(f"/v2/assets/{asset_id}/metadata")
        return cls._from_metadata(cls.session.get(url=url).json())

    @classmethod
    @utils.deprecation(None, "3.0.0")
    def all(
        cls,
        created_after: Optional[Union[str, dt.datetime]] = None,
        created_before: Optional[Union[str, dt.datetime]] = None,
        workspace_id: Optional[str] = None,
        collection_names: Optional[List[str]] = None,
        producer_names: Optional[List[str]] = None,
        tags: Optional[List[str]] = None,
        sources: Optional[List[str]] = None,
        search: Optional[str] = None,
        sort_by: Optional[utils.SortingField] = None,
    ) -> Iterator["Asset"]:
        params = {
            "createdAfter": created_after and utils.format_time(created_after),
            "createdBefore": created_before and utils.format_time(created_before),
            "workspaceId": workspace_id,
            "collectionNames": collection_names,
            "producerNames": producer_names,
            "tags": tags,
            "sources": sources,
            "search": search,
            "sort": sort_by,
        }
        return map(cls._from_metadata, utils.paged_query(params, "/v2/assets", cls.session))

    @property
    @utils.deprecation("pystac::Client.get_items", "3.0.0")
    @_retry
    def stac_items(self) -> pystac.ItemCollection:
        """Returns the stac items from an UP42 asset STAC representation."""
        try:
            filter_by_asset_id = {
                "max_items": MAX_ITEM,
                "limit": LIMIT,
                "filter": {
                    "op": "=",
                    "args": [
                        {"property": "asset_id"},
                        self.asset_id,
                    ],
                },
            }
            stac_search = self.stac_client.search(filter=filter_by_asset_id)
            return stac_search.item_collection()
        except Exception as exc:
            raise ValueError(f"No STAC metadata information available for this asset {self.asset_id}") from exc

    @property
    @utils.deprecation(None, "3.0.0")
    @_retry
    def stac_info(self) -> Union[pystac.Collection, pystac_client.CollectionClient]:
        """
        Gets the storage STAC information for the asset as a FeatureCollection.
        One asset can contain multiple STAC items (e.g. the PAN and multispectral images).
        """
        return self.stac_client.get_collection(self.stac_items[0].collection_id)

    @utils.deprecation("pystac::Item.update", "3.0.0")
    def update_metadata(
        self,
        title: Union[Optional[str], object] = NOT_PROVIDED,
        tags: Union[Optional[List[str]], object] = NOT_PROVIDED,
    ) -> dict:
        """
        Update the metadata of the asset.

        Args:
            title: The title string to be assigned to the asset. No value will keep the existing title.
            tags: A list of tag strings to be assigned to the asset. No value will keep the existing tags.

        Returns:
            The updated asset metadata information
        """
        if title != NOT_PROVIDED:
            self.title = cast(Optional[str], title)
        if tags != NOT_PROVIDED:
            self.tags = cast(Optional[list[str]], tags)
        self._save()
        return self.info

    def _save(self):
        url = host.endpoint(f"/v2/assets/{self.asset_id}/metadata")
        payload = {"title": self.title, "tags": self.tags}
        asset = self._from_metadata(self.session.post(url=url, json=payload).json())
        for field in dataclasses.fields(asset):
            setattr(self, field.name, getattr(asset, field.name))

    def _get_download_url(self, asset_id: str) -> str:
        url = host.endpoint(f"/v2/assets/{asset_id}/download-url")
        return self.session.post(url=url).json()["url"]

    @property
    @utils.deprecation("pystac::Asset.file", "3.0.0")
    def file(self) -> utils.ImageFile:
        return utils.ImageFile(url=self._get_download_url(self.asset_id))

    @utils.deprecation("pystac::Asset.file.url", "3.0.0")
    def get_stac_asset_url(self, stac_asset: pystac.Asset):
        """
        Returns the signed URL for the STAC Asset.
        Args:
            stac_asset: pystac Asset object.

        Returns:
            Signed URL for the STAC Asset.
        """
        stac_asset_id = stac_asset.href.split("/")[-1]
        return self._get_download_url(asset_id=stac_asset_id)

    @utils.deprecation("pystac::Asset.file::download", "3.0.0")
    def download(
        self,
        output_directory: Union[str, pathlib.Path, None] = None,
        unpacking: bool = True,
    ) -> List[str]:
        """
        Downloads the asset. Unpacking the downloaded file will happen as default.

        Args:
            output_directory: The file output directory, defaults to the current working
                directory.
            unpacking: By default the download TGZ/TAR or ZIP archive file will be unpacked.

        Returns:
            List of the downloaded asset filepaths.
        """
        logger.info("Downloading asset %s", self.asset_id)

        if output_directory is None:
            output_directory = pathlib.Path.cwd() / f"asset_{self.asset_id}"
        else:
            output_directory = pathlib.Path(output_directory)
        output_directory.mkdir(parents=True, exist_ok=True)
        logger.info("Download directory: %s", output_directory)

        download_url = self._get_download_url(asset_id=self.asset_id)
        download = utils.download_archive if unpacking else utils.download_file
        return download(
            download_url=download_url,
            output_directory=output_directory,
        )

    @utils.deprecation("pystac::Asset.file::download", "3.0.0")
    def download_stac_asset(
        self,
        stac_asset: pystac.Asset,
        output_directory: Union[str, pathlib.Path, None] = None,
    ) -> pathlib.Path:
        """
        Downloads a STAC asset to a specified output directory.

        Args:
            self (object): The instance of the class calling this method.
            stac_asset (pystac.Asset): The STAC asset to be downloaded.
            output_directory (Union[str, Path, None], optional): The directory where the asset
                will be downloaded. If not provided, a default directory structure will be used.

        Returns:
            Path: The path to the downloaded asset file.

        Raises:
            requests.exceptions.HTTPError: If there is an HTTP error during the download.

        """
        logger.info("Downloading STAC asset %s", stac_asset.title)
        if output_directory is None:
            output_directory = pathlib.Path.cwd() / f"asset_{self.asset_id}/{stac_asset.title}"
        else:
            output_directory = pathlib.Path(output_directory)
        output_directory.mkdir(parents=True, exist_ok=True)
        logger.info("Download directory: %s", output_directory)
        download_url = self.get_stac_asset_url(stac_asset=stac_asset)
        return pathlib.Path(utils.download_file(download_url, output_directory)[0])



import dataclasses as dc


@dc.dataclass(eq=True, frozen=True)
class ResilienceSettings:
    total: int = 5
    backoff_factor: float = 1
    statuses: tuple = tuple(range(500, 600))


@dc.dataclass(eq=True, frozen=True)
class TokenProviderSettings:
    token_url: str
    expiry_offset: int = 30
    timeout: int = 120


@dc.dataclass(eq=True, frozen=True)
class AccountCredentialsSettings:
    username: str
    password: str


CredentialsSettings = AccountCredentialsSettings



import functools
import pathlib
from typing import Callable, Dict, List, Optional, Union

import requests

from up42 import utils
from up42.http import config, oauth
from up42.http import session as http_session

SessionFactory = Callable[[oauth.Up42Auth], requests.Session]


class Client:
    def __init__(self, auth: oauth.Up42Auth, create_session: SessionFactory):
        self.auth = auth
        self.session = create_session(auth)


def _merge(
    left: Optional[config.CredentialsSettings],
    right: Optional[config.CredentialsSettings],
):
    if all([left, right]):
        raise MultipleCredentialsSources("Multiple sources of credentials provided")
    return left or right


SettingsDetector = Callable[[Optional[Dict]], Optional[config.CredentialsSettings]]
RetrieverDetector = Callable[[config.CredentialsSettings], oauth.TokenRetriever]
AuthFactory = Callable[[oauth.TokenRetriever, config.TokenProviderSettings], oauth.Up42Auth]


def create(
    credential_sources: List[Optional[Dict]],
    token_url: str,
    detect_settings: SettingsDetector = oauth.detect_settings,
    detect_retriever: RetrieverDetector = oauth.detect_retriever,
    create_auth: AuthFactory = oauth.Up42Auth,
    create_session: SessionFactory = http_session.create,
):
    possible_settings = [detect_settings(credentials) for credentials in credential_sources]
    settings = functools.reduce(_merge, possible_settings)
    if settings:
        token_settings = config.TokenProviderSettings(token_url=token_url)
        return Client(create_auth(detect_retriever(settings), token_settings), create_session)
    raise MissingCredentials


ConfigurationSource = Optional[Union[str, pathlib.Path]]
ConfigurationReader = Callable[[ConfigurationSource], Optional[Dict]]


def collect_credentials(
    cfg_file: ConfigurationSource,
    username: Optional[str],
    password: Optional[str],
    read_config: ConfigurationReader = utils.read_json,
) -> List[Optional[Dict]]:
    config_source = read_config(cfg_file)
    account_credentials_source = {"username": username, "password": password}
    return [config_source, account_credentials_source]


class MissingCredentials(ValueError):
    pass


class MultipleCredentialsSources(ValueError):
    pass



from typing import Optional, Union

import pystac

from up42 import base, host, utils


class InvalidUp42Asset(ValueError):
    pass


class FileProvider:
    session = base.Session()

    def __get__(self, obj: Optional[pystac.Asset], obj_type=None) -> Optional[utils.ImageFile]:
        if obj:
            if obj.href.startswith(host.endpoint("")):
                url = obj.href + "/download-url"
                signed_url = self.session.post(url=url).json()["url"]
                return utils.ImageFile(url=signed_url)
            else:
                return None
        else:
            raise AttributeError


UP42_USER_TITLE_KEY = "up42-user:title"
UP42_USER_TAGS_KEY = "up42-user:tags"


class UpdateItem:
    session = base.Session()

    def __call__(self, item):
        url = host.endpoint(f"/v2/assets/stac/collections/{item.collection_id}/items/{item.id}")
        response = pystac.Item.from_dict(
            self.session.patch(
                url=url,
                json={
                    UP42_USER_TITLE_KEY: item.properties.get(UP42_USER_TITLE_KEY),
                    UP42_USER_TAGS_KEY: item.properties.get(UP42_USER_TAGS_KEY),
                },
            ).json()
        )
        for key, value in vars(response).items():
            setattr(item, key, value)


class Up42ExtensionProperty:
    key: str

    def __init__(self, key):
        self.key = key

    def __get__(self, obj, obj_type=None):
        return obj._reference.get(self.key)

    def __set__(self, obj, value):
        obj._reference[self.key] = value


class Up42Extension:
    title = Up42ExtensionProperty("up42-user:title")
    tags = Up42ExtensionProperty("up42-user:tags")
    product_id = Up42ExtensionProperty("up42-product:product_id")
    collection_name = Up42ExtensionProperty("up42-product:collection_name")
    modality = Up42ExtensionProperty("up42-product:modality")
    order_id = Up42ExtensionProperty("up42-order:id")
    asset_id = Up42ExtensionProperty("up42-system:asset_id")
    account_id = Up42ExtensionProperty("up42-system:account_id")
    workspace_id = Up42ExtensionProperty("up42-system:workspace_id")
    job_id = Up42ExtensionProperty("up42-system:job_id")
    source = Up42ExtensionProperty("up42-system:source")
    metadata_version = Up42ExtensionProperty("up42-system:metadata_version")

    def __init__(self, reference: dict):
        self._reference = reference


class Up42ExtensionProvider:
    def __get__(self, obj: Optional[Union[pystac.Item, pystac.Collection]], obj_type=None):
        if obj:
            if obj_type == pystac.Item:
                return Up42Extension(obj.properties)  # type: ignore
            else:
                return Up42Extension(obj.extra_fields)  # type: ignore
        else:
            raise AttributeError


def extend():
    pystac.Asset.file = FileProvider()  # type: ignore

    update_item = UpdateItem()
    pystac.Item.update = lambda self: update_item(self)  # type: ignore # pylint: disable=unnecessary-lambda

    pystac.Item.up42 = Up42ExtensionProvider()  # type: ignore
    pystac.Collection.up42 = Up42ExtensionProvider()  # type: ignore



"""
Catalog search functionality
"""

import pathlib
from typing import Any, Dict, List, Literal, Optional, Union

import geojson  # type: ignore
import geopandas  # type: ignore
import tqdm
from shapely import geometry as geom  # type: ignore

from up42 import base, glossary, host, order, utils

logger = utils.get_logger(__name__)

Geometry = Union[
    dict,
    geojson.Feature,
    geojson.FeatureCollection,
    list,
    geopandas.GeoDataFrame,
    geom.Polygon,
]


class InvalidUsageType(ValueError):
    pass


class InvalidCollections(ValueError):
    pass


class MultipleHosts(ValueError):
    pass


class CatalogBase:
    """
    The base for Catalog and Tasking class, shared functionality.
    """

    session = base.Session()
    workspace_id = base.WorkspaceId()

    def __init__(self, collection_type: glossary.CollectionType):
        self.type = collection_type

    @utils.deprecation("DataProduct.schema", "3.0.0")
    def get_data_product_schema(self, data_product_id: str) -> dict:
        """
        Gets the parameters schema of a data product to help
        with the construction of the catalog/tasking order
        parameters.

        Args:
            data_product_id: The id of a catalog/tasking data product.
        """
        url = host.endpoint(f"/orders/schema/{data_product_id}")
        return self.session.get(url).json()

    @staticmethod
    @utils.deprecation("BatchOrderTemplate.estimate", "3.0.0")
    def estimate_order(order_parameters: order.OrderParams) -> int:
        """
        Estimate the cost of an order.

        Args:
            order_parameters: A dictionary like
            {dataProduct: ..., "params": {"id": ..., "aoi": ...}}
        Returns:
            int: An estimated cost for the order in UP42 credits.
        """

        return order.Order.estimate(order_parameters)

    @classmethod
    @utils.deprecation("BatchOrderTemplate::place", "3.0.0")
    def place_order(
        cls,
        order_parameters: order.OrderParams,
        track_status: bool = False,
        report_time: float = 120,
    ) -> order.Order:
        """
        Place an order.

        Args:
            order_parameters: A dictionary like
                {dataProduct: ..., "params": {"id": ..., "aoi": ...}}
            track_status (bool): If set to True, will only return
                the Order once it is `FULFILLED` or `FAILED`.
            report_time (int): The interval (in seconds) to query
                the order status if `track_status` is True.

         Warning:
            When placing orders of items that are in
            archive or cold storage,
            the order fulfillment can happen up to
            **24h after order placement**.
            In such cases, please make sure to set
            an appropriate `report_time`.
            You can also use `Order.track_status` on the
            returned object to track the status later.

        Returns:
            Order class object of the placed order.
        """
        placed_order = order.Order.place(order_parameters, cls.workspace_id)
        if track_status:
            placed_order.track_status(report_time)
        return placed_order


class Catalog(CatalogBase):
    """
    The Catalog class enables access to the UP42 catalog
    functionality (data archive search & ordering).

    Use catalog:
    ```python
    catalog = up42.initialize_catalog()
    ```

    This class also inherits functions from the
    [CatalogBase](catalogbase-reference.md) class.
    """

    def __init__(self):
        super().__init__(glossary.CollectionType.ARCHIVE)

    @staticmethod
    @utils.deprecation("Provider::search", "3.0.0")
    def construct_search_parameters(
        geometry: Geometry,
        collections: List[str],
        start_date: str = "2020-01-01",
        end_date: str = "2020-01-30",
        usage_type: Optional[List[Literal["DATA", "ANALYTICS"]]] = None,
        limit: int = 10,  # used for performance tuning and testing only
        max_cloudcover: Optional[int] = None,
    ) -> dict:
        """
        Helps constructing the parameters dictionary required for the search.

        Args:
            geometry: The search geometry, default a Polygon.
                One of FeatureCollection, Feature, dict (geojson geometry),
                list (bounds coordinates), GeoDataFrame,
                shapely.Polygon, shapely.Point.
                All assume EPSG 4326!
            collections: The satellite sensor collections to search for,
                e.g. ["phr"] or ["phr", "spot"].
                Also see catalog.get_collections().
            start_date: Query period starting day, format "2020-01-01".
            end_date: Query period ending day, format "2020-01-01".
            usage_type: Optional. Filter for imagery that can
                be purchased & downloaded or also processed.
                ["DATA"] (can only be downloaded),
                ["ANALYTICS"] (can be downloaded
                or used directly with a processing algorithm),
                ["DATA", "ANALYTICS"] (can be any combination).
                The filter is inclusive, using ["DATA"] can
                also result in results with ["DATA", "ANALYTICS"].
            limit: Page size in the API calls
            (used for performance tuning and testing only).
            max_cloudcover: Optional. Maximum cloud coverage percent.
                e.g. 100 will return all scenes,
                8.4 will return all scenes with 8.4 or less cloud coverage.
        Returns:
            The constructed parameters dictionary.
        """

        start = utils.format_time(start_date)
        end = utils.format_time(end_date, set_end_of_day=True)
        time_period = f"{start}/{end}"

        aoi_fc = utils.any_vector_to_fc(
            vector=geometry,
        )
        aoi_geometry = utils.fc_to_query_geometry(fc=aoi_fc, geometry_operation="intersects")

        query_filters: Dict[Any, Any] = {}
        if max_cloudcover is not None:
            query_filters["cloudCoverage"] = {"lte": max_cloudcover}

        if usage_type is not None:
            if set(usage_type) <= {"DATA", "ANALYTICS"}:
                query_filters["up42:usageType"] = {"in": usage_type}
            else:
                raise InvalidUsageType("usage_type is invalid")

        return {
            "datetime": time_period,
            "intersects": aoi_geometry,
            "limit": limit,
            "collections": collections,
            "query": query_filters,
        }

    def _get_host(self, collection_names: list[str]) -> str:
        collections = glossary.ProductGlossary.get_collections(collection_type=self.type)
        hosts = {
            provider.name
            for collection in collections
            if collection.name in collection_names
            for provider in collection.providers
            if "HOST" in provider.roles
        }

        if not hosts:
            raise InvalidCollections(
                f"Selected collections {collection_names} are not valid. See ProductGlossary.get_collections."
            )
        if len(hosts) > 1:
            raise MultipleHosts("Only collections with the same host can be searched at the same time.")
        return hosts.pop()

    @utils.deprecation("Provider::search", "3.0.0")
    def search(self, search_parameters: dict, as_dataframe: bool = True) -> Union[geopandas.GeoDataFrame, dict]:
        """
        Searches the catalog  and returns the metadata of the matching scenes.

        Args:
            search_parameters: The catalog search parameters, see example.
            as_dataframe: return type, geopandas.GeoDataFrame if True (default),
                FeatureCollection if False.

        Returns:
            The search results as a geopandas.GeoDataFrame,
            optionally as JSON dict.

        Example:
            ```python
                search_parameters={
                    "datetime": "2019-01-01T00:00:00Z/2019-01-15T23:59:59Z",
                    "collections": ["phr"],
                    "intersects": {
                        "type": "Polygon",
                        "coordinates": [[[13.32113746,52.73971768],
                        [13.15981158,52.2092959],[13.62204483,52.15632025],
                        [13.78859517,52.68655119],[13.32113746,
                        52.73971768]]]},
                    "limit": 10,
                    "sortby": [{"field" : "properties.acquisitionDate",
                        "direction" : "asc"}]
                    }
            ```
        """
        logger.info("Searching catalog with search_parameters: %s", search_parameters)

        if "limit" in search_parameters:
            search_parameters["limit"] = min(search_parameters["limit"], 500)

        product_host = self._get_host(search_parameters["collections"])
        url = host.endpoint(f"/catalog/hosts/{product_host}/stac/search")
        features = []

        while url:
            page: dict = self.session.post(url, json=search_parameters).json()
            features += page["features"]
            url = next((link["href"] for link in page["links"] if link["rel"] == "next"), None)

        if not features:
            df = geopandas.GeoDataFrame(columns=["geometry"], geometry="geometry")
        else:
            df = geopandas.GeoDataFrame.from_features(geojson.FeatureCollection(features=features), crs="EPSG:4326")

        logger.info("%s results returned.", df.shape[0])
        if as_dataframe:
            return df
        else:
            return df.__geo_interface__

    @utils.deprecation("BatchOrderTemplate", "3.0.0")
    def construct_order_parameters(
        self,
        data_product_id: str,
        image_id: str,
        aoi: Optional[Geometry] = None,
        tags: Optional[List[str]] = None,
    ) -> order.OrderParams:
        """
        Helps constructing the parameters dictionary required
        for the catalog order. Some collections have
        additional parameters that are added to the output
        dictionary with value None. The potential values to
        select from are given in the logs, for more detail on
        the parameter use `catalog.get_data_product_schema()`.

        Args:
            data_product_id: Id of the desired UP42 data product,
                see `catalog.get_data_products`
            image_id: The id of the desired image
                (from search results)
            aoi: The geometry of the order, one of dict, Feature,
                FeatureCollection, list, geopandas.GeoDataFrame, Polygon.
                Optional for "full-image products".
            tags: A list of tags that categorize the order.
        Returns:
            The order parameters dictionary.

        Example:
            ```python
            order_parameters = catalog.construct_order_parameters(
                data_product_id='647780db-5a06-4b61-b525-577a8b68bb54',
                image_id='6434e7af-2d41-4ded-a789-fb1b2447ac92',
                aoi={'type': 'Polygon',
                'coordinates': (((13.375966, 52.515068),
                  (13.375966, 52.516639),
                  (13.378314, 52.516639),
                  (13.378314, 52.515068),
                  (13.375966, 52.515068)),)})
            ```
        """
        schema = self.get_data_product_schema(data_product_id)
        params: dict[str, Any] = {param: None for param in schema["required"]}
        params["id"] = image_id
        order_parameters: order.OrderParams = {
            "dataProduct": data_product_id,
            "params": params,
        }
        if tags is not None:
            order_parameters["tags"] = tags

        # Some catalog orders, e.g. Capella don't require AOI (full image order)
        # Handled on API level, don't manipulate in SDK,
        # providers might accept geometries in the future.
        if aoi is not None:
            aoi = utils.any_vector_to_fc(vector=aoi)
            aoi = utils.fc_to_query_geometry(fc=aoi, geometry_operation="intersects")
            order_parameters["params"]["aoi"] = aoi  # type: ignore

        return order_parameters

    @utils.deprecation("Scene.quicklook::download", "3.0.0")
    def download_quicklooks(
        self,
        image_ids: List[str],
        collection: str,
        output_directory: Union[str, pathlib.Path, None] = None,
    ) -> List[str]:
        """
        Gets the quicklooks of scenes from a single sensor.

        Args:
            image_ids: List of provider image_ids
                e.g. ["6dffb8be-c2ab-46e3-9c1c-6958a54e4527"].
                Access the search results id column via
                `list(search_results.id)`.
            collection: The data collection corresponding to the image ids.
            output_directory: The file output directory,
                defaults to the current working directory.

        Returns:
            List of quicklook image output file paths.
        """
        product_host = self._get_host([collection])
        logger.info("Downloading quicklooks from provider %s.", product_host)

        if output_directory is None:
            output_directory = pathlib.Path.cwd() / "catalog"
        else:
            output_directory = pathlib.Path(output_directory)
        output_directory.mkdir(parents=True, exist_ok=True)
        logger.info("Download directory: %s", output_directory)

        out_paths: List[str] = []
        for image_id in tqdm.tqdm(image_ids):
            try:
                image_file = utils.ImageFile(
                    host.endpoint(f"/catalog/{product_host}/image/{image_id}/quicklook"),
                    f"quicklook_{image_id}.jpg",
                    session=self.session,
                )
                out_paths.append(str(image_file.download(output_directory)))
            except IOError:
                logger.warning(
                    "Image with id %s does not have quicklook available. Skipping ...",
                    image_id,
                )
        return out_paths
